#!/usr/bin/env python
# -*- coding: utf-8 -*-
# _version.py
"""Package initialization for DVHA-MLCA."""
# Copyright (c) 2021 Dan Cutright
# This file is part of dvha-mlca, released under a MIT license.
#    See the file LICENSE included with this distribution, also
#    available at https://github.com/cutright/DVHA-MLCA

__author__ = "Dan Cutright"
__email__ = "dan.cutright@gmail.com"
__version__ = "0.2.3"
__release__ = "0.2.3"
__version_info__ = (0, 2, 3)
